/*
 * Copyright 1995, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED
 *
 * This source code ("Source Code") was originally derived from a
 * code base owned by Silicon Graphics, Inc. ("SGI")
 * 
 * LICENSE: SGI grants the user ("Licensee") permission to reproduce,
 * distribute, and create derivative works from this Source Code,
 * provided that: (1) the user reproduces this entire notice within
 * both source and binary format redistributions and any accompanying
 * materials such as documentation in printed or electronic format;
 * (2) the Source Code is not to be used, or ported or modified for
 * use, except in conjunction with OpenGL Performer; and (3) the
 * names of Silicon Graphics, Inc.  and SGI may not be used in any
 * advertising or publicity relating to the Source Code without the
 * prior written permission of SGI.  No further license or permission
 * may be inferred or deemed or construed to exist with regard to the
 * Source Code or the code base of which it forms a part. All rights
 * not expressly granted are reserved.
 * 
 * This Source Code is provided to Licensee AS IS, without any
 * warranty of any kind, either express, implied, or statutory,
 * including, but not limited to, any warranty that the Source Code
 * will conform to specifications, any implied warranties of
 * merchantability, fitness for a particular purpose, and freedom
 * from infringement, and any warranty that the documentation will
 * conform to the program, or any warranty that the Source Code will
 * be error free.
 * 
 * IN NO EVENT WILL SGI BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT
 * LIMITED TO DIRECT, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES,
 * ARISING OUT OF, RESULTING FROM, OR IN ANY WAY CONNECTED WITH THE
 * SOURCE CODE, WHETHER OR NOT BASED UPON WARRANTY, CONTRACT, TORT OR
 * OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED BY PERSONS OR
 * PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED FROM,
 * OR AROSE OUT OF USE OR RESULTS FROM USE OF, OR LACK OF ABILITY TO
 * USE, THE SOURCE CODE.
 * 
 * Contact information:  Silicon Graphics, Inc., 
 * 1600 Amphitheatre Pkwy, Mountain View, CA  94043, 
 * or:  http://www.sgi.com
 */

/*
 * pfbyu.c: $Revision: 1.1 $ $Date: 2000/11/21 21:39:33 $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <Performer/pf.h>
#include <Performer/pfutil.h>
#include <Performer/pfdu.h>

/*
 * pfdLoadFile_byu -- Load ".byu" data files into IRIS Performer
 */

extern pfNode *
pfdLoadFile_byu (char *fileName)
{
    FILE	*byuFile	= NULL;
    pfNode	*node		= NULL;
    int		*partStart	= NULL;
    int		*partEnd	= NULL;
    pfVec3	*vertData	= NULL;
    pfdGeom	*prim		= NULL;
    int 	 i		= 0;
    int		 numParts	= 0;
    int		 numVerts	= 0;
    int		 numPolys	= 0;
    int		 numElements	= 0;
    int		 maxVerts	= 256;
    int		 inputPolys	= 0;
    pfMaterial  *m      	= NULL;
    double	 startTime	= pfGetTime();
    int		 t		= 0;

    /* restore builder to initial state */
    pfdResetBldrGeometry();
    pfdResetBldrState();

    /* open ".byu" file */
    if ((byuFile = pfdOpenFile(fileName)) == NULL)
	return NULL;

    /* allocate primitive buffer */
    prim = pfdNewGeom(maxVerts);

    /* pick a different set of colors each time */
    pfuRandomize(*(int *)&startTime);

    /* establish material definition for model */
    m = (pfMaterial*)pfdGetTemplateObject(pfGetMtlClassType());
    pfMtlColor(m, PFMTL_AMBIENT,  0.2f, 0.2f, 0.2f);
    pfMtlColor(m, PFMTL_DIFFUSE,  0.8f, 0.8f, 0.8f);
    pfMtlColor(m, PFMTL_SPECULAR, 1.0f, 1.0f, 1.0f);
    pfMtlShininess(m, 30.0f);
    pfMtlColorMode(m, PFMTL_FRONT, PFMTL_CMODE_AD);
    pfdBldrStateAttr(PFSTATE_FRONTMTL, m);
    pfdBldrStateMode(PFSTATE_ENLIGHTING, PF_ON);
    pfdBldrStateMode(PFSTATE_CULLFACE, PFCF_BACK);

    /* read count information from ".byu" file */
    if (fscanf(byuFile, "%d %d %d %d", 
	&numParts, 
	&numVerts, 
	&numPolys, 
	&numElements) != 4)
	return NULL;
    
    /* read part information from ".byu" file */
    partStart = (int *)calloc(numParts, sizeof(int));
    partEnd   = (int *)calloc(numParts, sizeof(int));
    for (i = 0; i < numParts; i++)
	fscanf(byuFile, "%d %d", 
	    &partStart[i], 
	    &partEnd[i]);

    /* read vertex information from ".byu" file */
    vertData = (pfVec3 *)calloc(numVerts, sizeof(pfVec3));
    for (i = 0; i < numVerts; i++)
	fscanf(byuFile, "%f %f %f", 
	    &vertData[i][0], 
	    &vertData[i][1], 
	    &vertData[i][2]);

    prim->nbind = PFGS_OFF;
    prim->cbind = PFGS_PER_PRIM;
    prim->primtype = PFGS_POLYS;
    for (t = 0 ; t < PF_MAX_TEXTURES ; t ++)
       prim->tbind[t] = PFGS_OFF;

    /* read polygon information from ".byu" file */
    for (i = 0; i < numParts; i++)
    {
	int	j;
	pfVec4	rgba;

	/* pick a random not-too-dark color */
	pfuRandomColor(rgba, 0.4f, 0.8f);

	/* read polygon definitions */
	for (j = partStart[i]; j <= partEnd[i]; j++)
	{
	    int k	 = 0;
	    int rawIndex = 0;

	    /* read polygon indices and build primitive */
	    while (rawIndex >= 0)
	    {
		int vertexIndex;

		/* read index from file */
		if (fscanf(byuFile, "%d", &rawIndex) != 1)
		{
		    pfNotify(PFNFY_WARN, PFNFY_PRINT, "pfdLoadFile_byu: premature end-of-file");
		    k = 0;
		    j = partEnd[i] + 1;
		    break;
		}

		/* adjust for 1-based indexing */
		vertexIndex = (rawIndex < 0) ? -rawIndex : rawIndex;
		vertexIndex--;

		/* check vertex */
		if (vertexIndex < 0 || vertexIndex >= numVerts)
		    pfNotify(PFNFY_WARN, PFNFY_PRINT, "pfdLoadFile_byu: bad vertex index");
		else
		{
		    /* enlarge primitive if needed */
		    if (k >= maxVerts)
			pfdResizeGeom(prim, maxVerts = 2*k);

		    /* dereference vertex array */
		    pfCopyVec3(prim->coords[k], vertData[vertexIndex]);

		    /* advance vertex count */
		    k++;
		}
	    }

	    if (k >= 3)
	    {
		/* set polygon color */
		pfCopyVec4(prim->colors[0], rgba);

		/* specify polygon control data */
		prim->numVerts = k;

		/* add this facet to builder */
		pfdAddBldrGeom(prim, 1);

		/* keep total counts */
		++inputPolys;
	    }
	}
    }

    /* close ".byu" file */
    fclose(byuFile);

    /* release primitive buffer */
    pfdDelGeom(prim);

    /* release dynamic data */
    free(partStart);
    free(partEnd);
    free(vertData);

    /* get a complete scene graph representing file's primitives */
    node = pfdBuild();

    /* use indicated name for top-level pfNode */
    if (node != NULL)
	pfNodeName(node, fileName);

    /* print statistics */
    pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_byu: %s", fileName);
    pfNotify(PFNFY_INFO,   PFNFY_MORE,  "  Input Data:");
    pfNotify(PFNFY_INFO,   PFNFY_MORE,  "    Input vertices:     %8ld", numVerts);
    pfNotify(PFNFY_INFO,   PFNFY_MORE,  "    Input polygons:     %8ld", inputPolys);

    return node;
}
