/*
 * Copyright 1993, 1995, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED
 *
 * This source code ("Source Code") was originally derived from a
 * code base owned by Silicon Graphics, Inc. ("SGI")
 * 
 * LICENSE: SGI grants the user ("Licensee") permission to reproduce,
 * distribute, and create derivative works from this Source Code,
 * provided that: (1) the user reproduces this entire notice within
 * both source and binary format redistributions and any accompanying
 * materials such as documentation in printed or electronic format;
 * (2) the Source Code is not to be used, or ported or modified for
 * use, except in conjunction with OpenGL Performer; and (3) the
 * names of Silicon Graphics, Inc.  and SGI may not be used in any
 * advertising or publicity relating to the Source Code without the
 * prior written permission of SGI.  No further license or permission
 * may be inferred or deemed or construed to exist with regard to the
 * Source Code or the code base of which it forms a part. All rights
 * not expressly granted are reserved.
 * 
 * This Source Code is provided to Licensee AS IS, without any
 * warranty of any kind, either express, implied, or statutory,
 * including, but not limited to, any warranty that the Source Code
 * will conform to specifications, any implied warranties of
 * merchantability, fitness for a particular purpose, and freedom
 * from infringement, and any warranty that the documentation will
 * conform to the program, or any warranty that the Source Code will
 * be error free.
 * 
 * IN NO EVENT WILL SGI BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT
 * LIMITED TO DIRECT, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES,
 * ARISING OUT OF, RESULTING FROM, OR IN ANY WAY CONNECTED WITH THE
 * SOURCE CODE, WHETHER OR NOT BASED UPON WARRANTY, CONTRACT, TORT OR
 * OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED BY PERSONS OR
 * PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED FROM,
 * OR AROSE OUT OF USE OR RESULTS FROM USE OF, OR LACK OF ABILITY TO
 * USE, THE SOURCE CODE.
 * 
 * Contact information:  Silicon Graphics, Inc., 
 * 1600 Amphitheatre Pkwy, Mountain View, CA  94043, 
 * or:  http://www.sgi.com
 */

/*
 * pfsgf.c: $Revision: 1.1 $ $Date: 2000/11/21 21:39:35 $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <Performer/pf.h>
#include <Performer/pfutil.h>
#include <Performer/pfdu.h>

#define	BUFFER_SIZE	256

/*
 * pfdLoadFile_sgf -- Load ".sgf" data files into IRIS Performer
 */

extern pfNode *
pfdLoadFile_sgf (char *fileName)
{
    FILE	*sgfFile	= NULL;
    pfNode	*node		= NULL;
    pfdGeom	*prim		= NULL;
    int		 numVerts	= 0;
    int		 maxVerts	= 0;
    int		 totalGroups	= 0;
    int		 totalVerts	= 0;
    double	 x		= 0.0;
    double	 y		= 0.0;
    double	 z		= 0.0;
    char	 buffer[BUFFER_SIZE];
    pfMaterial  *m      	= NULL;
    double	 startTime	= pfGetTime();
    int		 t		= 0;

    /* restore builder to initial state */
    pfdResetBldrGeometry();
    pfdResetBldrState();

    /* open ".sgf" file */
    if ((sgfFile = pfdOpenFile(fileName)) == NULL)
	return NULL;

    /* allocate primitive buffer */
    prim = pfdNewGeom(maxVerts = 256);

    /* pick a different set of colors each time */
    pfuRandomize(*(int *)&startTime);

    /* establish material definition for model */
    m = (pfMaterial*)pfdGetTemplateObject(pfGetMtlClassType());
    pfMtlColor(m, PFMTL_AMBIENT,  0.2f, 0.2f, 0.2f);
    pfMtlColor(m, PFMTL_DIFFUSE,  0.8f, 0.8f, 0.8f);
    pfMtlColor(m, PFMTL_SPECULAR, 1.0f, 1.0f, 1.0f);
    pfMtlShininess(m, 30.0f);
    pfMtlColorMode(m, PFMTL_FRONT, PFMTL_CMODE_AD);
    pfdBldrStateAttr(PFSTATE_FRONTMTL, m);
    pfdBldrStateMode(PFSTATE_ENLIGHTING, PF_ON);
    pfdBldrStateMode(PFSTATE_CULLFACE, PFCF_BACK);

    /* skip "descriptive string" */
    if (fgets(buffer, BUFFER_SIZE, sgfFile) == NULL)
    {
	fclose(sgfFile);
	return NULL;
    }

    /* pick a random not-too-dark color */
    pfuRandomColor(prim->colors[0], 0.4f, 0.8f);

    /* read lines from ".sgf" file */
    while (fgets(buffer, BUFFER_SIZE, sgfFile) != NULL)
    {
	/* enlarge primitive if needed */
	if (numVerts >= maxVerts)
	    pfdResizeGeom(prim, maxVerts = 2*numVerts);

	/* read vertex data */
	pfSetVec3(prim->coords[numVerts], 0.0f, 0.0f, 0.0f);
	sscanf(buffer, "%lf,%lf,%lf", &x, &y, &z);

	/* is this a data vertex? */
	if (x < 1.0e36 || y < 1.0e36 || z < 1.0e36)
	{
	    prim->coords[numVerts][PF_X] = x;
	    prim->coords[numVerts][PF_Y] = y;
	    prim->coords[numVerts][PF_Z] = z;
	    totalVerts++;
	    numVerts++;

	    if (numVerts > 1)
	    {
		/* specify control data */
		prim->numVerts = 2;
		prim->primtype = PFGS_LINES;
		prim->pixelsize = 1.0f;
		for (t = 0 ; t < PF_MAX_TEXTURES ; t ++)
		    prim->tbind[t] = PFGS_OFF;
		prim->nbind = PFGS_OFF;
		prim->cbind = PFGS_PER_PRIM;

		/* add this line to builder */
		pfdAddBldrGeom(prim, 1);

		/* save just-used vertex */
		pfCopyVec3(prim->coords[0], prim->coords[1]);
		numVerts = 1;
	    }
	}
	else
	/* is's a terminating vertex */
	{
	    /* pick a random not-too-dark color */
	    pfuRandomColor(prim->colors[0], 0.4f, 0.8f);

	    /* keep total counts */
	    totalGroups++;

	    /* reset vertex counter */
	    numVerts = 0;
	}
    }

    /* close ".sgf" file */
    fclose(sgfFile);

    /* release primitive buffer */
    pfdDelGeom(prim);

    /* get a complete scene graph representing file's primitives */
    node = pfdBuild();

    /* use indicated name for top-level pfNode */
    if (node != NULL)
	pfNodeName(node, fileName);

    /* print statistics */
    pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_sgf: %s", fileName);
    pfNotify(PFNFY_INFO,   PFNFY_MORE,  "  Input Data:");
    pfNotify(PFNFY_INFO,   PFNFY_MORE,  "    Input vertices:     %8ld", totalVerts);
    pfNotify(PFNFY_INFO,   PFNFY_MORE,  "    Input groups:       %8ld", totalGroups);

    return node;
}
