/*
 * Copyright 1995, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED
 *
 * This source code ("Source Code") was originally derived from a
 * code base owned by Silicon Graphics, Inc. ("SGI")
 * 
 * LICENSE: SGI grants the user ("Licensee") permission to reproduce,
 * distribute, and create derivative works from this Source Code,
 * provided that: (1) the user reproduces this entire notice within
 * both source and binary format redistributions and any accompanying
 * materials such as documentation in printed or electronic format;
 * (2) the Source Code is not to be used, or ported or modified for
 * use, except in conjunction with OpenGL Performer; and (3) the
 * names of Silicon Graphics, Inc.  and SGI may not be used in any
 * advertising or publicity relating to the Source Code without the
 * prior written permission of SGI.  No further license or permission
 * may be inferred or deemed or construed to exist with regard to the
 * Source Code or the code base of which it forms a part. All rights
 * not expressly granted are reserved.
 * 
 * This Source Code is provided to Licensee AS IS, without any
 * warranty of any kind, either express, implied, or statutory,
 * including, but not limited to, any warranty that the Source Code
 * will conform to specifications, any implied warranties of
 * merchantability, fitness for a particular purpose, and freedom
 * from infringement, and any warranty that the documentation will
 * conform to the program, or any warranty that the Source Code will
 * be error free.
 * 
 * IN NO EVENT WILL SGI BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT
 * LIMITED TO DIRECT, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES,
 * ARISING OUT OF, RESULTING FROM, OR IN ANY WAY CONNECTED WITH THE
 * SOURCE CODE, WHETHER OR NOT BASED UPON WARRANTY, CONTRACT, TORT OR
 * OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED BY PERSONS OR
 * PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED FROM,
 * OR AROSE OUT OF USE OR RESULTS FROM USE OF, OR LACK OF ABILITY TO
 * USE, THE SOURCE CODE.
 * 
 * Contact information:  Silicon Graphics, Inc., 
 * 1600 Amphitheatre Pkwy, Mountain View, CA  94043, 
 * or:  http://www.sgi.com
 *
 * culldl.c  code fragment for complex.c demonstrating use of cull 
 *	     callback for user rendering to display list.  
 *	     Uses MakeColorCube().
 *
 * $Revision: 1.1 $ $Date: 2000/11/21 21:39:36 $
 *
 */

#include <Performer/pf.h>

/* colorcube.c */
extern pfGeoSet* MakeColorCube(void);


pfMatrix Mat = {{ 1.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 1.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 1.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 1.0f }};

pfMatrix Xform = {{ 0.9998f,   0.02f, 0.0f, 0.0f },
		  {  -0.02f, 0.9998f, 0.0f, 0.0f },
		  {    0.0f,    0.0f, 1.0f, 0.0f },
		  {    0.0f,    0.0f, 0.0f, 1.0f }};

pfGeoSet *GSet = NULL;

/*
 *	CullChannel() -- traverse the scene graph and generate a display list
 *	for the draw process.
 */

static void
CullChannel(pfChannel *channel, void *data)
{
    pfCull();

    /* draw a geoset not in the scene hierarchy to the display list */

    /* make geoset if needed */
    if (GSet == NULL)
	GSet = MakeColorCube();

    /* compute new transform */
    pfPreMultMat(Mat, Xform);

    /* add a push matrix command to the display list */
    pfPushMatrix();

    /* add a mult matrix command to the display list */
    pfMultMatrix(Mat);

    /* add a draw geoset command to the display list */
    pfDrawGSet(GSet);

    /* add a pop matrix command to the display list */
    pfPopMatrix();
}
